use std::path::Path;

use tera::{Map, Value};

fn inner_merge<'a>(a: &mut Map<String, Value>, b: &'a Value) -> Option<Value> {
    match b {
        Value::Object(o) => {
            let mut map = Map::new();
            for (key, entry) in o.iter() {
                let mut child_map = if a.contains_key(key) {
                    a[key]
                        .as_object()
                        .map(|m| m.to_owned())
                        .unwrap_or(Map::new())
                } else {
                    Map::new()
                };

                if let Some(value) = inner_merge(&mut child_map, entry) {
                    child_map.insert(key.into(), value);
                }

                map.extend(child_map);
            }

            Some(map.into())
        }
        value => Some(value.to_owned()),
    }
}

trait Merge {
    fn merge(&mut self, other: Self);
}

impl Merge for tera::Context {
    fn merge(&mut self, other: Self) {
        let mut json = self.clone().into_json();
        let map = json.as_object_mut().unwrap();

        if let Some(value) = inner_merge(map, &other.into_json()) {
            if let Ok(new_map) = tera::Context::from_value(value) {
                self.extend(new_map)
            }
        }
    }
}

trait Head: Iterator {
    fn head(self) -> Option<(Self::Item, Self)>
    where
        Self: Sized;
}

impl<I: Iterator> Head for I {
    fn head(mut self) -> Option<(Self::Item, Self)> {
        match self.next() {
            Some(x) => Some((x, self)),
            None => None,
        }
    }
}

#[derive(Debug)]
struct MapFold<I, F, V> {
    iter: I,
    f: F,
    accum: V,
}

impl<I, F, V> MapFold<I, F, V> {
    pub fn new(iter: I, init: V, f: F) -> MapFold<I, F, V> {
        Self {
            iter,
            f,
            accum: init,
        }
    }
}

impl<I: Iterator, F: FnMut(&V, &I::Item) -> V, V: Clone> Iterator for MapFold<I, F, V> {
    type Item = V;

    fn next(&mut self) -> Option<Self::Item> {
        self.accum = (self.f)(&self.accum, &self.iter.next()?);
        Some(self.accum.clone())
    }
}

trait MapFoldExt {
    type Item;

    fn map_fold<B, F>(mut self, init: B, f: F) -> MapFold<Self, F, B>
    where
        Self: Sized,
        F: FnMut(&B, &Self::Item) -> B,
    {
        MapFold::new(self, init, f)
    }
}

impl<I: Iterator> MapFoldExt for I {
    type Item = I::Item;
}

#[derive(Debug)]
pub(crate) struct Context {
    inner: tera::Context,
}

impl Context {
    pub fn new() -> Self {
        Self {
            inner: tera::Context::new(),
        }
    }

    pub fn to_inner(&mut self) -> &mut tera::Context {
        &mut self.inner
    }

    fn path_to_string<P: AsRef<Path>>(path: &P) -> String {
        path.as_ref().to_string_lossy().to_string()
    }

    fn create_path<P: AsRef<Path>>(path: &P, value: &Value) -> Option<Value> {
        let path = path.as_ref();
        let (head, tail) = path.components().head()?;

        let mut map = Map::new();

        if tail.clone().count() > 0 {
            let child = Self::create_path(&tail, value)?;
            map.insert(Self::path_to_string(&head), child);
        } else {
            let key = Self::path_to_string(&path.with_extension("").file_name()?);
            map.insert(key, value.to_owned());
        }

        Some(map.into())
    }

    pub fn insert<P: AsRef<Path>>(&mut self, path: &P, value: &Value) {
        let path = path.as_ref();
        let path = if path
            .with_extension("")
            .file_name()
            .map_or(false, |f| f.to_os_string() == "index")
        {
            path.parent()
        } else {
            Some(path)
        };

        if let Some(v) = Self::create_path(&path.unwrap(), value) {
            if let Ok(ctx) = tera::Context::from_value(v) {
                self.inner.merge(ctx)
            }
        }
    }

    pub fn get<P: AsRef<Path>>(&self, path: &P) -> Option<&Value> {
        None
    }
}

#[cfg(test)]
mod tests {
    use std::path::Path;

    use super::Context;

    #[test]
    fn whatever() {
        let path = "this/is/a/test.html";
        let mut context = Context::new();
        println!("{context:?}");
        context.insert(&Path::new(&path), &":3".into());
        println!("{context:?}");
    }
}
